
from functools import cached_property
from typing import Callable, Dict, Union

from packaging import version

import abstracts

from aio.api.github import utils
from .base import GithubRepoEntity


class AGithubRelease(GithubRepoEntity, metaclass=abstracts.Abstraction):
    """A Github release."""

    def __str__(self):
        return f"<{self.__class__.__name__} {self.repo.name}@{self.tag_name}>"

    @cached_property
    def __data__(self) -> Dict[str, Callable]:
        return dict(
            created_at=utils.dt_from_js_isoformat,
            published_at=utils.dt_from_js_isoformat)

    @cached_property
    def version(self) -> Union[version.LegacyVersion, version.Version]:
        return version.parse(self.tag_name)



"""aio.api.github."""

from . import abstract
from . import exceptions
from . import utils
from .api import (
    GithubAPI,
    GithubCommit,
    GithubIssue,
    GithubIssues,
    GithubIterator,
    GithubLabel,
    GithubRelease,
    GithubRepo,
    GithubTag)
from .abstract import (
    AGithubAPI,
    AGithubCommit,
    AGithubIssue,
    AGithubIssues,
    AGithubIterator,
    AGithubLabel,
    AGithubRelease,
    AGithubRepo,
    AGithubTag)


__all__ = (
    "abstract",
    "AGithubAPI",
    "AGithubCommit",
    "AGithubIssue",
    "AGithubIssues",
    "AGithubIterator",
    "AGithubLabel",
    "AGithubRelease",
    "AGithubRepo",
    "AGithubTag",
    "exceptions",
    "GithubAPI",
    "GithubCommit",
    "GithubLabel",
    "GithubIssue",
    "GithubIssues",
    "GithubIterator",
    "GithubRelease",
    "GithubRepo",
    "GithubTag",
    "utils")




from typing import Any, Callable, Dict

from aio.api.github import abstract


UNSET = object()


class GithubEntity:
    """Base Github entity class."""

    def __init__(self, github: "abstract.AGithubAPI", data: Dict) -> None:
        self._github = github
        self.data = data

    @property
    def github(self) -> "abstract.AGithubAPI":
        """Github API."""
        return self._github

    @property
    def __data__(self) -> Dict[str, Callable]:
        """Dictionary of callables to mangle corresponding `self.data` keys."""
        return {}

    def __getattr__(self, k: str, default: Any = UNSET) -> Any:
        """Return the item from `self.data`, after mangling if required."""
        try:
            v = self.data[k]
        except KeyError:
            if default is not UNSET:
                return default
            return self.__getattribute__(k)
        return self.__data__.get(k, lambda x: x)(v)


class GithubRepoEntity(GithubEntity):
    """Base Github repo entity class."""

    def __init__(self, repo: "abstract.AGithubRepo", data: Dict) -> None:
        self.repo = repo
        self.data = data

    @property
    def github(self) -> "abstract.AGithubAPI":
        """Github API."""
        return self.repo.github




import pathlib
from datetime import datetime
from functools import cached_property, partial
from typing import Any, Dict, Optional, Type

import abstracts

from aio.api.github import abstract, exceptions, utils
from . import base


class AGithubRepo(metaclass=abstracts.Abstraction):
    """A Github repo."""

    def __init__(self, github: "abstract.AGithubAPI", name: str) -> None:
        self.github = github
        self.name = name

    def __str__(self):
        return f"<{self.__class__.__name__} {self.name}>"

    @cached_property
    def github_path(self) -> pathlib.PurePosixPath:
        """Github API repos path."""
        return pathlib.PurePosixPath(f"/repos/{self.name}")

    @cached_property
    def issues(self) -> "abstract.AGithubIssues":
        """Github issues for this repo."""
        return self.github.issues_class(
            self.github,
            repo=self)

    @property
    def labels(self) -> "abstract.AGithubIterator":
        """Github labels for this repo."""
        return self.iter_entities(
            self.github.label_class, "labels")

    async def commit(self, name: str) -> "abstract.AGithubCommit":
        """Fetch a commit for this repo."""
        return self.github.commit_class(
            self,
            await self.getitem(f"commits/{name}"))

    def commits(self, since: datetime = None) -> "abstract.AGithubIterator":
        """Iterate commits for this repo."""
        query = "commits"
        if since is not None:
            query = f"{query}?since={utils.dt_to_js_isoformat(since)}"
        return self.getiter(
            query,
            inflate=partial(self.github.commit_class, self))

    async def getitem(self, query: str) -> Any:
        """Call the `gidgethub.getitem` api for this repo."""
        return await self.github.getitem(self.github_endpoint(query))

    def getiter(self, query: str, **kwargs) -> "abstract.AGithubIterator":
        """Return a `GithubIterator` wrapping `gidgethub.getiter` for this
        repo."""
        return self.github.getiter(self.github_endpoint(query), **kwargs)

    def github_endpoint(self, rel_path: str) -> str:
        """Github API path for provided relative path."""
        return str(self.github_path.joinpath(rel_path))

    async def highest_release(
            self,
            since: Optional[datetime] = None) -> Optional[
                "abstract.AGithubRelease"]:
        """Release with the highest semantic version, optionally `since` a
        previous release date.

        Not necessarily the most recent.
        """
        highest_release = None

        async for release in self.releases():
            if since and release.published_at < since:
                break
            is_higher = (
                not release.prerelease
                and release.version
                and (not highest_release
                     or release.version >= highest_release.version))
            if is_higher:
                highest_release = release
        return highest_release

    def iter_entities(
            self,
            entity: Type[base.GithubEntity],
            path: str,
            **kwargs) -> "abstract.AGithubIterator":
        """Iterate and inflate entities for provided type."""
        return self.getiter(
            path,
            inflate=partial(entity, self),
            **kwargs)

    async def patch(self, query: str, data: Optional[Dict] = None) -> Any:
        """Call the `gidgethub.patch` api for this repo."""
        return await self.github.patch(self.github_endpoint(query), data=data)

    async def post(self, query: str, data: Optional[Dict] = None) -> Any:
        """Call the `gidgethub.post` api for this repo."""
        return await self.github.post(self.github_endpoint(query), data=data)

    async def release(self, name: str) -> "abstract.AGithubRelease":
        """Fetch a release for this repo."""
        return self.github.release_class(
            self,
            await self.getitem(f"releases/tags/{name}"))

    def releases(self) -> "abstract.AGithubIterator":
        """Iterate releases for this repo."""
        # TODO: make per_page configurable
        return self.iter_entities(
            self.github.release_class,
            "releases?per_page=100")

    async def tag(self, name: str) -> "abstract.AGithubTag":
        """Fetch a tag for this repo."""
        ref_tag = await self.getitem(f"git/ref/tags/{name}")
        if ref_tag["object"]["type"] != "tag":
            raise exceptions.TagNotFound(name)
        tag = await self.github.getitem(ref_tag["object"]["url"])
        return self.github.tag_class(self, tag)

    def tags(self) -> "abstract.AGithubIterator":
        """Iterate tags for this repo."""
        return self.iter_entities(
            self.github.tag_class,
            "tags")




from datetime import datetime
from functools import cached_property

import abstracts

from aio.api.github import utils
from .base import GithubRepoEntity


class AGithubCommit(GithubRepoEntity, metaclass=abstracts.Abstraction):
    """A Github commit."""

    def __str__(self):
        return f"<{self.__class__.__name__} {self.repo.name}#{self.sha}>"

    @cached_property
    def timestamp(self) -> datetime:
        """Datetime of the commit."""
        return utils.dt_from_js_isoformat(
            self.data["commit"]["committer"]["date"])




import urllib
from functools import cached_property, partial
from typing import Any, Callable, Dict

import gidgethub

import abstracts

from aio.api.github import abstract, exceptions

from .base import GithubRepoEntity


class AGithubIssue(GithubRepoEntity, metaclass=abstracts.Abstraction):
    """A Github issue."""

    def __gt__(self, other) -> bool:
        return self.number > other.number

    def __lt__(self, other) -> bool:
        return self.number < other.number

    def __str__(self):
        return f"<{self.__class__.__name__} {self.repo.name}#{self.number}>"

    async def close(self) -> "AGithubIssue":
        return await self.edit(state="closed")

    async def comment(self, comment: str) -> Any:
        """Add a comment to the issue."""
        # TODO: add comment class
        return await self.repo.post(
            f"issues/{self.number}/comments", data=dict(body=comment))

    async def edit(self, **kwargs) -> "AGithubIssue":
        """Edit the issue."""
        return self.__class__(
            self.repo,
            await self.repo.patch(f"issues/{self.number}", data=kwargs))


class AGithubIssues(metaclass=abstracts.Abstraction):
    """Github issues."""

    def __init__(
            self,
            github: "abstract.AGithubAPI",
            repo: "abstract.AGithubRepo" = None,
            filter: str = "") -> None:
        self.github = github
        self.repo = repo
        self._filter = filter

    @cached_property
    def filter(self) -> str:
        """Github search filter."""
        filter_parts = []
        if self._filter:
            filter_parts.append(self._filter)
        if self.repo:
            filter_parts.append(f"repo:{self.repo.name}")
        filters = " ".join(filter_parts)
        return filters and f"{filters} " or filters

    async def create(self, title: str, **kwargs) -> AGithubIssue:
        """Create an issue."""
        repo = kwargs.pop("repo", None) or self.repo
        if not repo:
            raise exceptions.IssueCreateError(
                f"To create an issue, either `{self.__class__.__name__}` "
                "must be instantiated with a `repo` or `create` must be "
                "called with one.")
        kwargs["title"] = title
        try:
            data = await repo.post("issues", data=kwargs)
        except gidgethub.GitHubException as e:
            raise exceptions.IssueCreateError(
                f"Failed to create issue '{title}' in {repo.name}\n"
                f"Recieved: {e}")
        return self.github.issue_class(repo, data)

    def inflater(self, repo: "abstract.AGithubRepo" = None) -> Callable:
        """Return default or custom callable to inflate a `GithubIssue`."""
        repo = repo or self.repo
        if not repo:
            return self._inflate
        return partial(self.github.issue_class, repo)

    def search(
            self,
            query: str,
            repo: "abstract.AGithubRepo" = None) -> "abstract.AGithubIterator":
        """Search for issues."""
        return self.github.getiter(
            self.search_query(query),
            inflate=self.inflater(repo))

    def search_query(self, query: str) -> str:
        """Generate a search query."""
        q = urllib.parse.quote(f"{self.filter}{query}")
        return f"/search/issues?q={q}"

    def _inflate(self, result: Dict) -> AGithubIssue:
        """Inflate an issue, finding the repo from the issue url."""
        return self.github.issue_class(
            self.github.repo_from_url(result["repository_url"]),
            result)




from typing import Any, AsyncGenerator, Dict, Mapping, Tuple

import gidgethub.abc
import gidgethub.sansio

import abstracts

from aio.core.functional import async_property


class AGithubIterator(metaclass=abstracts.Abstraction):
    """Async iterator to wrap gidgethub API and provide `total_count`."""

    def __init__(
            self,
            api: gidgethub.abc.GitHubAPI,
            query: str,
            *args,
            **kwargs) -> None:
        self.api = api
        self.query = query
        self.args = args
        self.kwargs = kwargs
        self._inflate = kwargs.pop("inflate", None)

    async def __aiter__(self) -> AsyncGenerator[Any, None]:
        """Async iterate an API call, inflating the results."""
        aiter = self.api.getiter(
            self.query, *self.args, **self.kwargs)
        async for item in aiter:
            yield self.inflate(item)

    @property
    def count_request_headers(self) -> Dict[str, str]:
        """Request headers for API call to get `total_count`."""
        request_headers = gidgethub.sansio.create_headers(
            self.api.requester,
            accept=gidgethub.sansio.accept_format(),
            oauth_token=self.api.oauth_token)
        request_headers["content-length"] = "0"
        return request_headers

    @property
    def count_url(self) -> str:
        """Request URL for API call to get `total_count`."""
        return gidgethub.sansio.format_url(
            f"{self.query}&per_page=1",
            {},
            base_url=self.api.base_url)

    @async_property(cache=True)
    async def total_count(self) -> int:
        """Get `total_count` without iterating all items."""
        # https://github.com/brettcannon/gidgethub/discussions/153
        if self.api.rate_limit is not None:
            self.api.rate_limit.remaining -= 1
        return self.count_from_response(
            await self.api._request(
                "GET", self.count_url, self.count_request_headers, b''))

    def count_from_data(self, data: Dict) -> int:
        """Get `total_count` from the data."""
        return (
            int(data["total_count"])
            if "total_count" in data
            else 0)

    def count_from_headers(self, headers: Mapping[str, str]) -> int:
        """Get the last page link from from the headers resulting from the
        request for single page iteration."""
        return (
            int(headers["Link"].split(
                ",")[1].split(
                    ">")[0].split("=")[-1])
            if "Link" in headers
            else 0)

    def count_from_response(
            self,
            response: Tuple[int, Mapping[str, str], bytes]) -> int:
        """Get total count from the headers or data."""

        (data,
         self.api.rate_limit,
         more) = gidgethub.sansio.decipher_response(*response)

        return (
            self.count_from_headers(response[1])
            if "Link" in response[1]
            else self.count_from_data(data))

    def inflate(self, result: Any) -> Any:
        """Inflate a result."""
        return (
            self._inflate(result)
            if self._inflate
            else result)




import abstracts

from .base import GithubRepoEntity


class AGithubLabel(GithubRepoEntity, metaclass=abstracts.Abstraction):
    """A Github label."""

    def __str__(self):
        return f"<{self.__class__.__name__} {self.name}>"




import abc
from functools import cached_property
from typing import Any, Type

import aiohttp

import gidgethub
import gidgethub.aiohttp

import abstracts

from .commit import AGithubCommit
from .issues import AGithubIssue, AGithubIssues
from .iterator import AGithubIterator
from .label import AGithubLabel
from .release import AGithubRelease
from .repo import AGithubRepo
from .tag import AGithubTag


class AGithubAPI(metaclass=abstracts.Abstraction):
    """Github API wrapper.

    Can be used to access the gidgethub API directly.

    Can also be used to work with repos by calling `self["repo/name"]`.
    """

    def __init__(
            self,
            session: aiohttp.ClientSession,
            *args, **kwargs) -> None:
        self.session = session
        self.args = args
        self.kwargs = kwargs

    def __getitem__(self, k) -> AGithubRepo:
        """Return a `GithubRepository` for `k`"""
        # TODO: make this work with user and organization
        #  and validate `k`
        return self.repo_class(self, k)

    @cached_property
    def api(self) -> gidgethub.aiohttp.GitHubAPI:
        """Gidgethub API."""
        return self.api_class(
            self.session,
            *self.args,
            **self.kwargs)

    @property
    @abc.abstractmethod
    def api_class(self) -> Type[gidgethub.aiohttp.GitHubAPI]:
        """API class."""
        return gidgethub.aiohttp.GitHubAPI

    @property  # type:ignore
    @abstracts.interfacemethod
    def commit_class(self) -> Type[AGithubCommit]:
        """Github commit class."""
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    def label_class(self) -> Type[AGithubLabel]:
        """Github label class."""
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    def issue_class(self) -> Type[AGithubIssue]:
        """Github issue class."""
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    def issues_class(self) -> Type[AGithubIssues]:
        """Github issues class."""
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    def iterator_class(self) -> Type[AGithubIterator]:
        """Github iterator class.

        Provides both an async iterator and a `total_count` async prop.
        """
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    def release_class(self) -> Type[AGithubRelease]:
        """Github release class."""
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    def repo_class(self) -> Type[AGithubRepo]:
        """Github repo class."""
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    def tag_class(self) -> Type[AGithubTag]:
        """Github tag class."""
        raise NotImplementedError

    async def getitem(self, *args, **kwargs) -> Any:
        """Call the `gidgethub.getitem` api."""
        # print("GETITEM", args, kwargs)
        return await self.api.getitem(*args, **kwargs)

    def getiter(self, *args, **kwargs) -> AGithubIterator:
        """Return a `GithubIterator` wrapping `gidgethub.getiter`."""
        # print("GETITER", args, kwargs)
        return self.iterator_class(self.api, *args, **kwargs)

    async def patch(self, *args, **kwargs):
        """Call the `gidgethub.patch` api."""
        # print("PATCH", args, kwargs)
        return await self.api.patch(*args, **kwargs)

    async def post(self, *args, **kwargs):
        """Call the `gidgethub.post` api."""
        # print("POST", args, kwargs)
        return await self.api.post(*args, **kwargs)

    def repo_from_url(self, url):
        """Return the corresponding `GithubRepo` for an api url."""
        repo_url = f"{self.api.base_url}/repos/"
        if not url.startswith(repo_url):
            return None
        return self["/".join(url[len(repo_url):].split("/")[:2])]




from .api import AGithubAPI
from .commit import AGithubCommit
from .issues import AGithubIssue, AGithubIssues
from .iterator import AGithubIterator
from .label import AGithubLabel
from .release import AGithubRelease
from .repo import AGithubRepo
from .tag import AGithubTag


__all__ = (
    "AGithubAPI",
    "AGithubCommit",
    "AGithubIssue",
    "AGithubIssues",
    "AGithubIterator",
    "AGithubLabel",
    "AGithubRelease",
    "AGithubRepo",
    "AGithubTag")




import abstracts

from aio.api.github import abstract
from aio.core.functional import async_property
from .base import GithubRepoEntity


class AGithubTag(GithubRepoEntity, metaclass=abstracts.Abstraction):

    def __str__(self):
        return f"<{self.__class__.__name__} {self.repo.name}@{self.tag}>"

    @async_property(cache=True)
    async def commit(self) -> "abstract.AGithubCommit":
        """Related commit for this tag."""
        return await self.repo.commit(self.object["sha"])




from datetime import datetime


# these only deal with utc but are good enough for working with the
# github api


def dt_from_js_isoformat(iso: str) -> datetime:
    return datetime.fromisoformat(iso.replace("Z", "+00:00"))


def dt_to_js_isoformat(dt: datetime) -> str:
    return dt.isoformat().replace("+00:00", "Z")




from typing import Type

import gidgethub.aiohttp

import abstracts

from .abstract import (
    AGithubAPI,
    AGithubCommit,
    AGithubIssue,
    AGithubIssues,
    AGithubIterator,
    AGithubLabel,
    AGithubRelease,
    AGithubRepo,
    AGithubTag)


@abstracts.implementer(AGithubCommit)
class GithubCommit:
    pass


@abstracts.implementer(AGithubIssue)
class GithubIssue:
    pass


@abstracts.implementer(AGithubIssues)
class GithubIssues:
    pass


@abstracts.implementer(AGithubIterator)
class GithubIterator:
    pass


@abstracts.implementer(AGithubLabel)
class GithubLabel:
    pass


@abstracts.implementer(AGithubRelease)
class GithubRelease:
    pass


@abstracts.implementer(AGithubRepo)
class GithubRepo:
    pass


@abstracts.implementer(AGithubTag)
class GithubTag:
    pass


@abstracts.implementer(AGithubAPI)
class GithubAPI:

    def __init__(self, session, *args, **kwargs) -> None:
        AGithubAPI.__init__(self, session, *args, **kwargs)

    @property
    def api_class(self) -> Type[gidgethub.aiohttp.GitHubAPI]:
        return super().api_class

    @property
    def commit_class(self) -> Type[AGithubCommit]:
        return GithubCommit

    @property
    def issue_class(self) -> Type[AGithubIssue]:
        return GithubIssue

    @property
    def issues_class(self) -> Type[AGithubIssues]:
        return GithubIssues

    @property
    def iterator_class(self) -> Type[AGithubIterator]:
        return GithubIterator

    @property
    def label_class(self) -> Type[AGithubLabel]:
        return GithubLabel

    @property
    def release_class(self) -> Type[AGithubRelease]:
        return GithubRelease

    @property
    def repo_class(self) -> Type[AGithubRepo]:
        return GithubRepo

    @property
    def tag_class(self) -> Type[AGithubTag]:
        return GithubTag
